from .scla import scla, datasets
from .__about__ import __version__, __title__
from .exceptions import ExtractError
import sys
import argparse


def scla_script():
    parser = argparse.ArgumentParser(
        description='scla {} ({} variant) - Analyse sound latencies'.format(__version__, __title__),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--runid', '-i', help='An id for the sound plot', default='scla')
    parser.add_argument('soundfile', help='A sound recording of port and sound output')
    parser.add_argument('logfile', help='Stimulus delivery recording of event sequence')
    parser.add_argument('--schannel', '-c', help='The channel sounds were recorded in', default=1, type=int)
    parser.add_argument('--maxdur', '-m', help='Max duration for sound/port event', default=0.012, type=float)
    parser.add_argument('--thresh', '-t', help='Threshold for sound/port detection', default=0.2, type=float)
    parser.add_argument('--version', action='version', version='scla {} ({} variant)'.format(__version__, __title__))
    parser.add_argument('--precision', '-p', help='Limit the precision of reported values to this many decimal places', default=3, type=int)
    parser.add_argument('--results', '-r', help='Limit the results to only fields of interest', choices=datasets, default=list(datasets.keys()), nargs='+')
    parser.add_argument('--plot', help='Plot the first sound to a png file detected for visual inspection - requires matplotlib to be installed', action='store_true')

    cmdargs = vars(parser.parse_args())
    precision = cmdargs.pop('precision')
    reports = cmdargs.pop('results')

    try:
        res = scla(**cmdargs)
    except ExtractError as e:
        print(e)
        sys.exit(65)
    report = ['==================================']
    for result in res:
        if result not in reports:
            continue
        report.append(result)
        for measure in ['mean', 'min', 'max', 'stddev']:
            report.append('\t{0}: {1:.{2}f}'.format(measure, res[result][measure], precision))
        report.append('----------------------------------')
    report.append(report[0])

    print('\n'.join(report))



class LoadError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class DataNotFoundError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class ExtractError(Exception):
    """Raised when an extraction goes wrong"""
    def __init__(self, logData, portData, sndData):
        super(ExtractError, self).__init__()
        self.logData, self.portData, self.sndData = logData, portData, sndData
        self.message = 'Extraction error, detected code lengths do not match'

    def __str__(self):
        """Represent this error as a string"""
        elements = [self.message, len(self.logData), len(self.portData), len(self.sndData)]
        return '{}\nLengths: Log({}), Port({}), Snds({})'.format(*elements)


class ConvertError(Exception):
    """Raised when a log file number was not converted successfully"""
    def __init__(self, udat):
        super().__init__()
        self.udat = udat


class SoundError(Exception):
    """raised when there is an issue loading the soundfile"""
    def __init__(self, msg):
        super().__init__()
        self.msg = msg



from __future__ import division
from .exceptions import LoadError
from datetime import datetime
from itertools import takewhile
import os


class Experiment:
    def __init__(self, name):
        self.name = name


class Measure:
    '''An object that represents some manipulation or subset of the data in
    question'''
    def __init__(self, name):
        self.name = name

    def build(self):
        """Should be set by user code, build a subset of a data object
        and return it"""
        raise NotImplementedError('Build should be defined by user code')


class Event:
    def __init__(self, bline, header):
        """Create Event info from line and header dict"""
        self.data = {h: (bline[header[h]] if header[h] < len(bline) else '')
                     for h in header}
        self.etype = self.data['Event Type']
        self.code = self.data['Code']
        self.time = float(self.data['Time']) / 10  # In ms

    def __str__(self):
        return self.etype + ':' + self.code + ':' + str(self.time)

    def __repr__(self):
        return str(self)

    @property
    def time_sec(self):
        return self.time / 1000


class Record:
    def __init__(self, filename):
        if os.path.isfile(filename):
            self.source = os.path.abspath(filename)
            self._extract()
        else:
            raise LoadError('Err: Logfile not found: {}'.format(filename))

    def _extract(self):
        """Extract usable data from the file into the record dictionary
        """
        with open(self.source, 'r') as lf:
            lines = lf.read().splitlines()
        l = 0  # A pointer which we'll use to step through the file
        # Grab Scenario name
        if lines[l].split('-')[0].strip() == 'Scenario':
            self.exp = Experiment(lines[l].split('-')[1].strip())
            l += 1
        else:
            raise LoadError('Err: first line does not start with Scenario')
        # Get Logfile timestamp
        if lines[l].split('-')[0].strip() == 'Logfile written':
            self.timestamp = datetime.strptime(lines[l].split('-')[1].strip(),
                                               '%m/%d/%Y %H:%M:%S')
            l += 1
        else:
            raise LoadError('Err: second line does not contain timestamp')
        # Grab Logfile Header
        for line in lines[l:]:
            if line.startswith('Subject') or line.startswith('Trial'):
                hline = line.split('\t')
                unc_types = [' (Duration)', ' (Time)']
                hline = [h + unc_types.pop() if h == 'Uncertainty' else h
                         for h in hline]
                self.header = {hline[c]: c for c in range(len(hline))}
                l += 1
                break
            else:
                l += 1
        if lines[l] != '':
            raise LoadError('Err: Expected blank line between header and body')
        # Grab data
        l += 1
        self.events = []
        for line in takewhile(lambda x: len(x) > 0, lines[l:]):
            self.events.append(Event(line.split('\t'), self.header))
        try:
            self.subjectID = self.events[0].data['Subject']
        except KeyError:
            self.subjectID = ''

    def segment(self, smarker, emarker):
        self.segments = []
        inblock = False
        for item in self.events:
            if item.code == smarker:
                inblock = True
                self.segments.append([])
            elif item.code == emarker:
                if not inblock:
                    print('Warning: found end block before start block')
                inblock = False
            elif inblock:
                self.segments[-1].append(item)


def load(f):
    '''Load a file and create the resulting Record object'''
    return Record(f)


def subset(n, builder):
    '''Create a Measure with name n and builder function'''
    m = Measure(n)
    m.build = builder
    return m



# coding: utf-8
from __future__ import division
import statistics as stat
from .logfile import load
from .exceptions import ExtractError, SoundError
from copy import deepcopy
import wave
import struct

datasets = {'Port_to_Port': None, 'Snd_to_Snd': None, 'Port_Length': None, 'Snd_Upper_Bound': [], 'Snd_Lower_Bound': []}


def stdStats(datasets):
    """Run the full gamut of standard stats on each dataset"""
    stats = {}
    for d in datasets:
        stats[d] = {}
        data = datasets[d]
        stats[d]['mean'] = stat.mean(data)
        stats[d]['min'] = min(data)
        stats[d]['max'] = max(data)
        stats[d]['stddev'] = stat.stdev(data)
        stats[d]['rawdata'] = data
    return stats


def extract_channel_events(channel, maxdur=0.012, thresh=0.2, samplerate=44100):
    '''From a single channel, extract events that reach *thresh*old and last
    for *maxdur*'''
    events = []
    lastevt = -20000
    for index, value in enumerate(channel):
        if index - lastevt > maxdur * samplerate:
            if abs(value) > thresh:
                events.append(index / samplerate)
                lastevt = index
    return events


def extract_sound_events(soundfile, schannel=1, maxdur=0.012, thresh=0.2, plot=False, runid='scla'):
    fs, sdata = wavLoad(soundfile)  # wavfile.read(soundfile)
    # Grab channels
    port = sdata[1 - schannel]
    snd = sdata[schannel]
    # Normalise channels
    maxp = max(port)
    maxs = max(snd)
    port = [p / maxp for p in port]
    snd = [s / maxs for s in snd]
    port_events = extract_channel_events(port, maxdur=maxdur, thresh=thresh, samplerate=fs)
    snd_events = extract_channel_events(snd, maxdur=maxdur, thresh=thresh, samplerate=fs)
    if plot:
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            print('matplotlib not found, skipping plot of first sound')
        plt.plot(snd[int(snd_events[0] * fs) - 100:int((snd_events[0] + maxdur) * fs)])
        plt.savefig(runid + '_firstsnd.png')
        plt.close()
    return fs, port_events, snd_events, port


def scla(soundfile=None, logfile=None, **kwargs):
    """Implements similar logic to Neurobehavioural Systems SCLA program"""
    log = load(logfile)

    fs, pcodes, snds, port = extract_sound_events(soundfile, **kwargs)
    if (len(log.events) != len(pcodes)) or (len(pcodes) != len(snds)):
        raise ExtractError(log.events, pcodes, snds)

    thisdata = deepcopy(datasets)
    for evt in range(len(snds)):
        # Uncertainty in seconds
        uncertainty = float(log.events[evt].data['Uncertainty (Time)']) * 0.0001
        thisdata['Snd_Lower_Bound'].append(snds[evt] - pcodes[evt])
        thisdata['Snd_Upper_Bound'].append(snds[evt] - pcodes[evt] + uncertainty)
    td, pl = timing(port, pcodes, snds, fs, **kwargs)
    thisdata['Port_to_Port'] = td['pcodes']
    thisdata['Snd_to_Snd'] = td['snds']
    thisdata['Port_Length'] = pl
    return stdStats(thisdata)


def timing(port, pcodes, snds, fs, maxdur=0, thresh=0, **kwargs):
    """extract extra info about port duration and time between stimuli"""
    timediffs = {'pcodes': [], 'snds': []}
    portlengths = []
    for p in range(1, len(pcodes)):
        timediffs['pcodes'].append(pcodes[p] - pcodes[p - 1])
    for s in range(1, len(snds)):
        timediffs['snds'].append(snds[s] - snds[s - 1])
    for p in pcodes:
        span = range(int(p * fs) + 1, int((p + maxdur * 1000) * fs))
        for pt in span:
            if port[pt] < thresh:
                portlengths.append(pt / fs - p)
                break
    return timediffs, portlengths


def wavLoad(fname):
    '''Load a 2 channel wavefile into a list of two element lists
    Taken from: http://stackoverflow.com/a/2602334/1468125'''
    wav = wave.open(fname, "r")
    (nchannels, sampwidth, framerate, nframes, comptype, compname) = wav.getparams()
    frames = wav.readframes(nframes * nchannels)
    out = struct.unpack_from("%dh" % nframes * nchannels, frames)

    if nchannels == 2:
        return framerate, (out[0::2], out[1::2])  # ((L, R) channels)
    else:
        raise SoundError('Should be two and only two channels in recorded sound')
